/*
 *
 *	xbaff.c		-- But Another Folder Flasher (for X)
 *
 *	
 *	This program runs in the background looking at the user's  mail  spool
 *	file.  When it detects that there has been a change to the file, or at
 *	periodical intervals when there is outstanding mail for the user.
 *	
 *	Originally created for the pixrect world, this  is  being  changed  to
 *	work  with  the X world. It flashes up a window on the screen with the
 *	names of the senders, and the subject matter, for new email that comes
 *	into the spool file.
 *	
 *	The  general  architecture  of this is *not* all that good. It doesn't
 *	follow  the general event based structure of X, and leans toward being
 *	synchronous. Just a hack, as it were.
 *	
 *	To build:	  cc -o xbaff xbaff.c -lX11  -- or Imakefile
 *
 *	Jim Becker	  Spring91 X version created.
 *
 *	Jim Becker	  5/17/91  added color, removed stipple
 *			           added window location
 *	Jim Becker	  5/22/91  fixed bell and parse bugs
 *	Jim Becker	  7/12/91  added in lots of X junk
 *				   for WM, bitmaps, resources.
 *	Jim Becker	  7/19/91  gargling, leds and other
 *				   refinements suggested by lemke
 *	Jim Becker	  8/1/91   removed use of SizeHints on msg window
 *
 *
 * 	Copyright (C) 1991 Network Computing Devices
 *
 *	Permission to use, copy, modify, distribute, and  sell  this  software
 *	and its documentation for any purpose is hereby granted  without  fee,
 *	provided that the above copyright notice appear in all copies and that
 *	both that copyright  notice  and  this  permission  notice  appear  in
 *	supporting documentation, and  that  the  name  of  Network  Computing
 *	Devices  not  be  used  in  advertising  or  publicity  pertaining  to
 *	distribution  of  the  software  without   specific,   written   prior
 *	permission.   Network Computing Devices makes no representations about
 *	the suitability of this software for any purpose.  It is provided  "as
 *	is" without express or implied warranty.
 *	
 *	NETWORK COMPUTING DEVICES DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *	SOFTWARE, INCLUDING ALL  IMPLIED  WARRANTIES  OF  MERCHANTABILITY  AND
 *	FITNESS, IN NO EVENT SHALL NETWORK COMPUTING DEVICES BE LIABLE FOR ANY
 *	SPECIAL,  INDIRECT  OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
 *	RESULTING  FROM  LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
 *	CONTRACT,  NEGLIGENCE  OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 *	CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
#include	<ctype.h>
#ifdef SYSV
#include	<string.h>
#else
#include	<strings.h>
#endif
#include	<stdio.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include 	<signal.h>

#include 	<X11/X.h>
#include 	<X11/Xlib.h>
#include	<X11/Xatom.h>
#include	<X11/Xutil.h>

#define	APPLIC_NAME	"xbaff"		/* but another folder flasher! 	*/
#define	CLASS_NAME	"XBaff"		/* but another Folder Flasher! 	*/
#define	PRG_VERSION	"V2.00"		/* program version		*/
	
#define	DELAY_TIME	60		/* time between checks		*/
#define	THRESHOLD	60		/* #times before show it	*/
#define	WAIT_TIME	2		/* how long on screen		*/
#define	BORDER_WIDTH	4		/* width of the screen border	*/
#define	DEF_KEYLED	0		/* led to light, zero means none*/
#define	COMFORT		2		/* min distance from borders	*/
#define	FONT_NAME	"9x15"		/* default font to use		*/
#define	LOC_NAME	"tc"		/* default location for message */
#define	DEF_FORE	"black"		/* default foreground color	*/
#define	DEF_BACK	"white"		/* default background color	*/
#define	DEF_GEOMETRY	"=+520+0"	/* where to place the icon win	*/
#define	RING_BELL	TRUE		/* ring the bell by default	*/
#define	GARGLE_PATHS	FALSE		/* try to rethink mail addresses*/

#define	STR_MAX		256
#define	MAX_LETTERS	50
#define	MAX_SAVE	80
#define MAX_XLATE	100

#define	TRUE		1
#define	FALSE		0

#define	EQUALN(a,b)	(strncmp((char*)a,(char*)b,strlen(a))==0)
#define	EQUAL(a,b)	(strcmp((char*)a,(char*)b)==0)
#define	EventMask	(ExposureMask)

#define	ArePendingEvents()	(XPending(display) != 0)
#define	NoPendingEvents()	(XPending(display) == 0)

typedef	struct {
	char		user[MAX_SAVE];
	char		subject[MAX_SAVE];
}	letter;

typedef	enum	{
	TopCenter,
	NorthEast,
	NorthWest,
	SouthEast,
	SouthWest
} Location;

/*	here are all the totally ridiculous things		*/
/*	that every application has to tell window manager	*/
/*	for it to be a "legal" citizen. how about just SS#?	*/
static	char		xbaff_prg_name[]	= APPLIC_NAME;
static	char		xbaff_class_name[]	= CLASS_NAME;
static	char		xbaff_title_name[]	= APPLIC_NAME;
static	char		xbaff_icon_name[]	= APPLIC_NAME;

/*	the following can be changed by the user	*/
static	int		delay_time;
static	int		threshhold;
static	int		wait_time;
static	int		border_width;
static	short		do_bell;			/* ring bell */
static	short		gargle_paths;			/* hack mail paths */

static	char		*font_name;
static	int		fwidth, fheight;

static	letter		letter_stack[MAX_LETTERS];
static	int		letter_count, old_letter_count;
static	int		max_user_len, max_subject_len;

static 	char		*xlate_name[MAX_XLATE];
static	char		num_xlate;

static	char		file_name[STR_MAX];
static  char            def_loc_name[STR_MAX];
static	int		last_file_time, current_file_time;
static	short		user_active;
static	short		subject_active;


static	int		dwidth, dheight;		/* display */
static	int		cwidth, cheight;		/* current */
static	Location	location;

static	Display		*display;
static	int		screen;
static	char		*dpy_name	= NULL;
static	Window		RootWin;
static	Window		window, mwindow;
static	short		is_mono;

static	int		forec, backc, bordc;
static	char		*forec_name, *backc_name, *bordc_name;
static	char		*bitmap_name;
static	char		*loc_name;
static	GC		regular, shadow, backgc;
static	Font		fid;
static	XGCValues	gcvalues;
static	int		gcvaluemask;
static	XSizeHints	xsize;
static	char		*geometry;
static	int		keyboard_led;

static	Pixmap		mail_pixmap;
static	Pixmap		icon_pixmap;
static	int		mailx, maily;
static	short		is_mail;

static	XEvent		xevent;

static	int		mail_width 	= 64;
static	int		mail_height 	= 48;
static	int		mail_hotX, mail_hotY;
static char mail_bits[] = {
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x03, 0x00, 0xff, 0xff, 0xff, 0xff,
   0xff, 0xff, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00, 0x60, 0x00, 0x03, 0x00,
   0xfb, 0x0f, 0x00, 0x00, 0xfe, 0x7c, 0x03, 0x00, 0x03, 0x00, 0x00, 0x80,
   0x61, 0x46, 0x03, 0x00, 0xfb, 0x0f, 0x00, 0x40, 0x7e, 0x6f, 0x03, 0x00,
   0x03, 0x00, 0x00, 0x80, 0x61, 0x61, 0x03, 0x00, 0xfb, 0x0f, 0x00, 0x40,
   0x7e, 0x73, 0x03, 0x00, 0xf3, 0x07, 0x00, 0x80, 0x61, 0x7e, 0x03, 0x00,
   0x03, 0x00, 0x00, 0x40, 0xfe, 0x18, 0x03, 0x00, 0x03, 0x1e, 0x00, 0x80,
   0x61, 0x18, 0x03, 0x00, 0x03, 0x00, 0x00, 0x40, 0x60, 0x3c, 0x03, 0x00,
   0x03, 0x00, 0x00, 0x00, 0x60, 0x7e, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00,
   0xe0, 0x81, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00, 0xe0, 0xff, 0x03, 0x00,
   0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00,
   0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00,
   0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x03, 0x00, 0x77, 0xff,
   0x01, 0x00, 0x03, 0x00, 0x03, 0x00, 0xdd, 0x6d, 0x03, 0x00, 0x03, 0x00,
   0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x03, 0x00, 0xf3, 0xf6,
   0x02, 0x00, 0x03, 0x00, 0x03, 0x00, 0xde, 0x9d, 0x03, 0x00, 0x03, 0x00,
   0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x03, 0x00, 0xf7, 0xef,
   0x03, 0x00, 0x03, 0x00, 0x03, 0x00, 0x5d, 0xbb, 0x02, 0x00, 0x03, 0x00,
   0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x03, 0x00, 0x03, 0x00, 0x00, 0x60, 0x1b, 0x00, 0x03, 0x00,
   0x03, 0x00, 0x00, 0xe0, 0x1f, 0x00, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00,
   0xc3, 0x01, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x23, 0x02, 0x00, 0x00,
   0x00, 0x00, 0x03, 0x00, 0xd3, 0x04, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00,
   0x93, 0x04, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0xd3, 0x05, 0x00, 0x00,
   0x00, 0x00, 0x03, 0x00, 0x23, 0x02, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00,
   0xc3, 0x01, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x03, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00,
   0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0xff, 0xff, 0xff, 0xff,
   0xff, 0xff, 0x03, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x03, 0x00};

#define stipple_width 4
#define stipple_height 4
static char stipple_bits[] = {
   0x01, 0x02, 0x04, 0x08};


static	char		*usage_msg[]	= {
    "\n         xbaff ",PRG_VERSION,
    "   - \"But Another Folder Flasher (for X)\"\n\n",

    "This  program  runs in the background looking at the user's mail spool\n",
    "file. When it detects that there has been a change to the file, or  at\n",
    "periodical  intervals  when there is outstanding mail for the user, it\n",
    "alerts the user on screen with the message and sender information. \n\n",

    "There are tweakable things, of course, with command line options:\n\n",

    "    -display  <host>     the display to use when running\n",
    "    -delay    <nn>       delay for <nn> seconds between mail checks\n",
    "    -thresh   <nn>       every <nn> times checked display old info\n",
    "    -wait     <nn>       wait for <nn> seconds per line before erasing info\n",
    "    -location <location> relative location of notification window\n",
    "    -quiet               quiet on new mail rather than ring bell\n",
    "    -geometry <geometry> x and y offsets for mail icon (std format)\n",
    "    -fg       <color>    named color for message window foreground\n",
    "    -bg       <color>    named color for message window background\n",
    "    -bd       <color>    named color for message window border\n",
    "    -bw       <nn>       width of the border on the message window\n",
    "    -led      <nn>       led of keyboard to indicate mail waiting\n",
    "    -bitmap   <bitmap>   bitmap to use for mail notification icon\n",
    "    -font     <fontname> font to use for text in message window\n",
    "    -gargle              gargle the mailpath names to simplify\n",
    "    -h                   this (mondo verbose) help text output\n",
    "    -v                   current xbaff program version output\n\n",

    "The  location window for xbaff by default appears in the top center of\n",
    "the  user's  display.  This  can  be  adjusted  by  specifying compass\n",
    "quadrants of the form:\n\n",

    "     tc     -- top center of the screen (default)\n",
    "     ne     -- northeast for the upper right corner\n",
    "     nw     -- northwest for the upper left  corner\n",
    "     se     -- southeast for the lower right corner\n",
    "     sw     -- southwest for the lower left  corner\n\n",

    "This stuff is all available in resources as well, here's an example:\n\n",

    "    xbaff.delayTime:        1\n",
    "    xbaff.threshold:        10\n",
    "    xbaff.waitTime:         3\n",
    "    xbaff.borderWidth:      10\n",
    "    xbaff.bitmapName:       /u/jcb/bitmaps/smallmoon.xbm\n",
    "    xbaff.geometry:         =-30-30\n",
    "    xbaff.background:       brown\n",
    "    xbaff.foreground:       red\n",
    "    xbaff.border:           blue\n",
    "    xbaff.mailLed:          3\n",
    "    xbaff.location:         nw\n",
    "    xbaff.ringBell:         True\n",
    "    xbaff.garglePaths:      True\n",
    "    xbaff.fontName:         9x15\n\n",

    "There  is  an associated file that xbaff uses to simplify sender names\n",
    "displayed. This is useful if the sender has a long email  address  and\n",
    "one  just  wants  to  see  the   username.   This   file   is   called\n",
    "$HOME/.xbaff_xlate and has the following format:\n\n",

    "	<actual_email_address>~<logical_name>\n\n",

    "So the following file will result in 'Bugs Bunny' being used  for  the\n",
    "name. Note that the match string is for an exact match only, including\n",
    "embedded white space. Enable gargling if you want xbaff to guess..\n\n",

    "uunet!so-cal!disney!sunglasses!bb.rabbit~Bugs Bunny\n\n",

    "Default settings are:\n\n",
    "xbaff    -delay 60 -thresh 60 -wait 2 -loc tc -geometry =+520+0 -font 9x15\n",
    "xadvised -delay 60 -thresh 60 -wait 2 -loc se -geometry =+520+0 -font 9x15\n\n",

    NULL};

extern	char		*getenv();

mail_file_time()
{
struct	stat		file_stat;

	if( stat( file_name, &file_stat) != 0 )
		return -1;
	else
		return (int)file_stat.st_mtime;
}

/*
 *	this is used to trim down a username into something that look more 'normal'.
 *	lot of potential pitfalls here, as all the names can be of different formats.
 */
char	*
get_simple_user( userp )
char	*userp;
{
	static	char	username[STR_MAX];
	char		*p;
	int		i, j, len;
	char		*left, *right;
	short		found	= FALSE;
	
	len	= strlen( userp );

	bzero( username, sizeof(username) );

	for( i = 0; i < num_xlate && !found; i++ ) {

		if( strncmp( xlate_name[i], userp, len ) == 0 ) {

			p	= xlate_name[i] + len;

			if( *p++ == '~' ) {
				strcpy( username, p );
				found	= TRUE;
			}
		}
	}

	if( found )
		return username;

	/* the code below trys to gargle the names to something more reasonable */
	if( !gargle_paths ) {
		strcpy( username, userp );
		return username;
	}

	/* if we haven't found a match, try to simplify the name by looking for
	 * the address, as <address>, and stripping. if that doesn't work, look
	 * for double quotes, followed by parens. finally, if all else fails
	 * remove the address starting with a bang ("!"). 
	 */

	left	= strchr(  userp, (int)('<') );
	right	= strrchr( userp, (int)('>') );

	/* copy over string without path in arrows */
	if( left && right && left < right ) {

		for( j = 0, p = userp; *p ; p++ ) {
			if( p < left || p > right )
				username[j++] = *p;
		}
		goto final_trim;
	}

	left	= strchr(  userp, (int)('\"') );
	right	= strrchr( userp, (int)('\"') );

	if( left && right && left != right ) {
		strncpy( username, ++left, right-left-1 );
		goto final_trim;
	}

	left	= strchr(  userp, (int)('(') );
	right	= strrchr( userp, (int)(')') );

	if( left && right && left < right ) {
		strncpy( username, ++left, right-left-1 );
		goto final_trim;
	}

	/* if there's a bang ("!") in it, remove what leads to it.. */
	right	= strrchr( userp, (int)('!') );

	if( right ) {
		strcpy( username, ++right );
		goto final_trim;
	}

	/* just copy it over verbatim, and see what happens */
	strcpy( username, userp );

final_trim:

	/* remove trailing space if there is one.. */
	len	= strlen(username);
	while( len > 0 && username[len-1] == ' ' ) len--;

	username[len]	= '\0';
	
	return username;
}

add_current_record()
{
	int		len;
	
	len	= strlen(letter_stack[letter_count].user);

	if( len > max_user_len )
		max_user_len	= len;

	len	= strlen(letter_stack[letter_count].subject);

	if( len > max_subject_len )
		max_subject_len	= len;

	letter_count++;

	user_active	= FALSE;
	subject_active	= FALSE;	
}		

clear_current_record()
{
	user_active	= FALSE;
	subject_active	= FALSE;	

	if( letter_count > 0 )
		letter_count--;
}		

set_current_user( user )
char		*user;
{
	char		*p	= get_simple_user(user);

	/* trim if too long */
	if( strlen(p) >= MAX_SAVE )
		p[MAX_SAVE-1] = '\0';

	strcpy( letter_stack[letter_count].user, p );

	user_active	= TRUE;
}

set_current_subject( subject )
char		*subject;
{
	char		*p	= subject;

	/* zap leading spaces */
	while( *p == ' ' )
		p++;

	/* trim if too long */
	if( strlen(p) >= MAX_SAVE )
		p[MAX_SAVE-1] = '\0';

	strcpy( letter_stack[letter_count].subject, p );

	subject_active	= TRUE;
}

open_parse_file()
{
	FILE		*mail;
	char		line[STR_MAX];
	short		found_header	= FALSE;
	short		done		= FALSE;

	mail	= fopen( file_name, "r" );

	old_letter_count = letter_count;

	letter_count	= 0;
	max_user_len	= 0;
	max_subject_len	= 0;
	clear_current_record();

	if( mail == NULL )
		return letter_count;

	bzero( line, sizeof(line) );
	
	while( !done && fgets( line, sizeof(line)-1, mail ) != NULL ) {

		line[strlen(line)-1] = '\0';

		/* scan for the next "From " sequence */
		while( !done && !found_header ) {

			found_header = EQUALN( "From ", line );

			if( !found_header ) {
				if( fgets( line, sizeof(line), mail ) == NULL )
					done	= TRUE;

				line[strlen(line)-1] = '\0';
			}
		}

		if( EQUALN( "From:", 	line ) )
			set_current_user( &line[6] );
		else
		if( EQUALN( "Subject:", line ) )
			set_current_subject( &line[9] );
		else
		if( EQUALN( "Status:", 	line ) )
			clear_current_record();

		if( user_active && subject_active ) {
			add_current_record();
			found_header	= FALSE;

			/* insure there isn't a rollover of the array */
			if( letter_count == MAX_LETTERS-2 ) {
				set_current_user(    "** more mail **"      );
				set_current_subject( "<<read your mail!!>>" );
				add_current_record();	

				done = TRUE;
			}
		}

		/* paranoid clearing! */
		bzero( line, sizeof(line) );
	}						

	fclose(mail);

	return letter_count;
}

/*
 *	write a single line to the display
 */
update_line( lineno )
int	lineno;
{
	char		*p;
	int		x, y, i;

	/* too many calls to the same routine here... */

	/* put out the username */

        i       = letter_count - lineno - 1;
	p	= letter_stack[i].user;
	x	= fwidth;
	y	= (lineno+1) * fheight + 5;

	/* create the name - in pseudo-bold */
	XDrawString( display, window, regular, x-1, y, p, strlen(p)); 
	XDrawString( display, window, regular, x,   y, p, strlen(p)); 

	/* put out the subject */

	p	=  letter_stack[i].subject;
	x	= (max_user_len + 4) * fwidth;

	XDrawString( display, window, regular, x,   y, p, strlen(p));
}

/*
 *	update the information that the user gets to see.
 */
update_display()
{
	int		i;

	reposition_x_window();

	if( do_bell && letter_count > old_letter_count ) 
		XBell( display, 50 );

	XMapRaised( display, window );

	wait_for_expose( window );

	XFillRectangle( display, window, backgc, 0, 0, cwidth, cheight );

	for( i = 0; i < letter_count; i++ ) 
		update_line( i );

	XFlush( display );

	wait_then_unmap();

}

/*
 *	this spits out the message on how to use the demo program.
 */
static	void
usage()
{
	char		**string 	= usage_msg;

	while( *string != NULL )
		printf( *string++ );
}

/*
 *	compares two strings for equality, with case folding
 */
extern	Bool
strequal( stra, strb )
char		*stra, *strb;
{
	int		ca, cb;

	/* check length before anything */
	if( strlen(stra) != strlen(strb) )
		return False;

	do {
		ca	= *stra++;
		cb	= *strb++;

		if( islower(ca) )
			ca	= toupper(ca);

		if( islower(cb) )
			cb	= toupper(cb);
	}
	while( ca == cb && ca != 0 );

	return (ca == 0);
}

/*
 *	load the defaults from the resource file
 */
static	char*
def_get_default( opt, defstr )
char		*opt;
char		*defstr;
{
	char		*retstr;
	
	retstr	= (char *)XGetDefault( display, xbaff_prg_name, opt );

	return (retstr ? retstr : defstr);
}


static	Bool
def_get_boolean( boolname, booldef )
char		*boolname;
Bool		booldef;
{
	char		*boolstr;

	boolstr	= def_get_default( boolname, (booldef ? "True" : "False" ));

	/* determine what the string means */
	if( strequal( boolstr, "True" ) )
		return True;
	else if( strequal( boolstr, "False") )
		return False;
	else if( strequal( boolstr, "On") )
		return True;
	else if( strequal( boolstr, "Off") )
		return False;
	else
		return booldef;
}

static	int
def_get_integer( intname, intdef )
char		*intname;
Bool		intdef;
{
	char		*intstr;

	intstr	= def_get_default( intname, "<none>");

	/* determine what the string means */
	if( strequal( intstr, "<none>" ) )
		return intdef;
	else
		return atoi(intstr);
}


/*
 *	load all the defaults from Xdefaults stuff first
 */
load_resource_defaults()
{
	delay_time	= def_get_integer( "delayTime",   DELAY_TIME   	);
	threshhold	= def_get_integer( "threshold",   THRESHOLD    	);
	wait_time	= def_get_integer( "waitTime",    WAIT_TIME    	);
	border_width	= def_get_integer( "borderWidth", BORDER_WIDTH	);
	keyboard_led	= def_get_integer( "mailLed",     DEF_KEYLED	);
	
	bitmap_name	= def_get_default( "bitmapName",  "<none>" 	);
	geometry	= def_get_default( "geometry", 	  DEF_GEOMETRY 	);
	backc_name	= def_get_default( "background",  DEF_BACK  	);
	forec_name	= def_get_default( "foreground",  DEF_FORE 	);
	bordc_name	= def_get_default( "border",	  DEF_FORE 	);
	font_name	= def_get_default( "fontName", 	  FONT_NAME 	);
	loc_name	= def_get_default( "location",    def_loc_name 	);

	gargle_paths	= def_get_boolean( "garglePaths", GARGLE_PATHS	);
	do_bell		= def_get_boolean( "ringBell",	  RING_BELL	);
}


/*
 *	parse for the -display option
 */
parse_display_name( argc, argv )
int		argc;
char		**argv;
{
	char		*arg;
	int		i;

	for( i = 1; i < argc; i++ ) {

		arg	= argv[i];

		if( strequal( arg, "-display" ) ) {
			if( i < (argc-1) ) {
				dpy_name	= argv[++i];
				return;
			}
		}
	}
}
	       

/*
 *	parse command line arguments
 */
parse_args( argc, argv )
int		argc;
char		**argv;
{
	char		*arg;
	int		i;
	short		got_error	= FALSE;

	for( i = 1; i < argc; i++ ) {

		arg	= argv[i];

		if( arg[0] == '-' ) {
			switch( arg[1] ) {
			case 	't':
				if( i < (argc-1) ) {
					threshhold	= atoi(argv[++i]);
				}
				break;
			case 	'w':
				if( i < (argc-1) ) {
					wait_time	= atoi(argv[++i]);
				}
				break;
			case	'd':
				if( i < (argc-1) ) {
					switch( arg[2] ) {
					case 'i':
						dpy_name	= argv[++i];
						break;
					case 'e':
						delay_time	= atoi(argv[++i]);
						break;
					}
				}
				break;
			case	'f':
				if( i < (argc-1) ) {
					switch( arg[2] ) {
					case 'o':
						font_name	= argv[++i];
						break;
					case 'g':
						forec_name	= argv[++i];
						break;
					}
				}
				break;
			case	'l':
				if( i < (argc-1) ) {
					loc_name	= argv[++i];
				}
				break;
			case	'b':
				if( i < (argc-1) ) {
					switch( arg[2] ) {
					case 'g':
						backc_name	= argv[++i];
						break;
					case 'i':
						bitmap_name	= argv[++i];
						break;
					case 'w':
						border_width	= atoi(argv[++i]);
						break;
					case 'd':
						bordc_name	= argv[++i];
						break;
					}
				}
				break;
			case	'g':
				if( i < (argc-1) ) {
					switch( arg[2] ) {
					case 'e':
						geometry	= argv[++i];
						break;
					}
				}
				if( arg[2] == 'a' )
					gargle_paths	= True;
				break;
				
			case	'q':
				do_bell	= FALSE;
				break;
			case 	'-':
			case	'h':
			case	'H':
			case	'?':
				usage();
				got_error 	= TRUE;
				break;
			case	'v':
			case	'V':
				printf("xbaff %s\n", PRG_VERSION );
				got_error	= TRUE;
				break;
			default:
				printf("don't understand argument `%s'\n", arg);
			}
		}
	}

	return !got_error;
}

/*
 *	determine the location user has specified.
 */
set_location()
{
	if( EQUAL( loc_name, "tc" ) )
		location	= TopCenter;
	else
	if( EQUAL( loc_name, "ne" ) )
		location	= NorthEast;
	else
	if( EQUAL( loc_name, "nw" ) )
		location	= NorthWest;
	else
	if( EQUAL( loc_name, "se" ) )
		location	= SouthEast;
	else
	if( EQUAL( loc_name, "sw" ) )
		location	= SouthWest;
}

	
/*
 *	==== all the X stuff is done below here ====
 */

/*
 *	return allocated color or default color passed.
 */
get_color( name, pdefault )
char		*name;
int		pdefault;
{
	Colormap	cmap	= DefaultColormap(display, screen);
	XColor  	hard_color;
	XColor  	exact_color;

	if( name == NULL || strlen(name) == 0 || is_mono )
		return pdefault;

	if( XAllocNamedColor( display, cmap, name, &hard_color, &exact_color) )
		return hard_color.pixel;
	else
		return pdefault;
}
	

static	int
init_x_world( argc, argv )
int		argc;
char		**argv;
{
	XSetWindowAttributes 	attributes;
	XFontStruct		*fontinf;
	XWMHints		wmhints;
	XTextProperty 		tp, ip;
	XClassHint 		ch;
	long			valuemask;
	Pixmap			stipple;

	/* see if there's a display name */
	parse_display_name( argc, argv );

	display	= XOpenDisplay( dpy_name );

	if( display == NULL ) {
		printf("** Can't access the display !! **\n");
		return FALSE;
	}

	/* load the resource defaults */
	load_resource_defaults();

	/* parse the command line options */
	if( !parse_args( argc, argv ) )
		return FALSE;

	screen	= DefaultScreen(display);
	RootWin	= RootWindow(display, screen );

	is_mono	= (DisplayPlanes( display, screen ) < 2);
	backc	= get_color( backc_name, WhitePixel(display,screen ));
	forec	= get_color( forec_name, BlackPixel(display,screen ));
	bordc	= get_color( bordc_name, BlackPixel(display,screen ));

	dwidth 	= DisplayWidth(display,screen );
	dheight	= DisplayHeight(display,screen );


	fontinf	= XLoadQueryFont( display, font_name );

	if( fontinf == NULL ) {
		printf("** Can't open specified font '%s' fatal !! **\n", font_name );
		return FALSE;
	}

	fwidth	= fontinf->max_bounds.width;
	fheight	= fontinf->max_bounds.ascent + fontinf->max_bounds.descent + 3;
	fid	= fontinf->fid;
	
	attributes.background_pixel	= backc;
	attributes.border_pixel 	= bordc;
	attributes.override_redirect	= True;

	valuemask			= CWBackPixel 		| 
					  CWBorderPixel 	| 
					  CWOverrideRedirect;
	/* create message window */
	window	= XCreateWindow( display, RootWin, 
			    0, 0, 100, 100, border_width,
			    0, InputOutput, CopyFromParent, 
			    valuemask,
			    &attributes );

	if( window == NULL ) {
		printf("** Can't open display window !! **\n");
		return FALSE;
	}

	XSelectInput( display, window, EventMask  );

	/* create backdrop stipple */
	stipple	= XCreateBitmapFromData( display, window, 
					stipple_bits, 
					stipple_width, stipple_height );

	if( stipple == NULL ) {
		printf( "Error  stipple for background!\n");
		return FALSE;
	}

	/* setup gc for rendering logos to screen */
	gcvalues.foreground	= forec;
	gcvalues.background	= backc;
	gcvalues.font		= fid;
	gcvalues.function	= GXcopy;

	gcvaluemask		= GCFunction 	| GCForeground 	| 
				  GCBackground	| GCFont;

	/* create normal render gc for logo rendering */
	regular	= XCreateGC( display, window, gcvaluemask, &gcvalues );

	if( regular == NULL ) {
		printf("Error allocating render graphics context!\n");
		return FALSE;
	}

	gcvalues.foreground	= backc;

	shadow	= XCreateGC( display, window, gcvaluemask, &gcvalues );

	if( shadow == NULL ) {
		printf("Error allocating shadow graphics context!\n");
		return FALSE;
	}

	/* create gc used just to clear out the window to the background */
	gcvalues.foreground	= backc;
	gcvalues.stipple	= stipple;
	gcvalues.fill_style	= FillStippled;

	gcvaluemask		|= GCFillStyle	| GCStipple;

	backgc	= XCreateGC( display, window, gcvaluemask, &gcvalues );

	if( backgc == NULL ) {
		printf("Error allocating inversion graphics context!\n");
		return FALSE;
	}


	icon_pixmap	= XCreateBitmapFromData( display, window,
						mail_bits, mail_width, mail_height );

	mail_width	= 50;	/* kludge to convert 64 wide to 50 in icon */

	/* has the user specified a name for the render bitmap? */
	if( *bitmap_name && !strequal( bitmap_name, "<none>") ) {
		if( XReadBitmapFile( display, window, bitmap_name,
				(unsigned int *)&mail_width,(unsigned int * )&mail_height, 
				&mail_pixmap,
				&mail_hotX, &mail_hotY ) ) 
			printf("Error loading bitmap file `%s' - use default\n",
			       bitmap_name);
	}	
	
	/* second chance if bitmap load error */
	if( mail_pixmap == NULL )
		mail_pixmap	= icon_pixmap;
	
	/* still problems?!? */
	if( mail_pixmap == NULL ) {
		printf("Error allocating the mail pixmap!\n");
		return FALSE;
	}

	attributes.override_redirect	= False;

	/* add in the backing pixmap */
/*	attributes.background_pixmap	= mail_pixmap;
	valuemask		       |= CWBackPixmap;
*/

	/* this is the control window -- and the icon */
	mwindow	= XCreateWindow( display, RootWin, 
			    0, 0, mail_width, mail_height, 1,
			    0, InputOutput, CopyFromParent, 
			    valuemask, &attributes );

	if( mwindow == NULL ) {
		printf("** Can't open mail icon display window !! **\n");
		return FALSE;
	}

	XSelectInput( display, mwindow, EventMask  );

	/* parse the geometry if the user specified it */
	determine_icon_geometry();

	xsize.x			= mailx;
	xsize.y			= maily;
	xsize.width		= mail_width;
	xsize.height		= mail_height;
	xsize.min_width		= mail_width;
	xsize.min_height	= mail_height;
	xsize.max_width		= mail_width;
	xsize.max_height	= mail_height;

	xsize.flags		= USPosition | USSize |
				  PPosition  | PSize  |
				  PMinSize   | PMaxSize;

	/* set up all the junk for the WM properties */
	ch.res_name 		= xbaff_prg_name;
	ch.res_class 		= xbaff_class_name;

	tp.value 		= (unsigned char *)xbaff_title_name;
	tp.encoding 		= XA_STRING;
	tp.format 		= 8;
	tp.nitems 		= strlen(xbaff_title_name);

	ip.value 		= (unsigned char *)xbaff_icon_name;
	ip.encoding 		= XA_STRING;
	ip.format 		= 8;
	ip.nitems 		= strlen(xbaff_icon_name);

	wmhints.flags		= InputHint | StateHint | 
		                  IconPixmapHint | IconMaskHint;
	wmhints.input		= False;
	wmhints.initial_state	= NormalState;
	wmhints.icon_pixmap	= icon_pixmap;
	wmhints.icon_mask	= icon_pixmap;

	XSetWMProperties( display, mwindow, &tp, &ip, argv, argc, 
			 &xsize, &wmhints, &ch );

	XSetCommand( display, mwindow, argv, argc );
	XMoveResizeWindow(display, mwindow, mailx, maily, mail_width, mail_height );

	/* set location of message window */
	set_location();

	return TRUE;
}

determine_icon_geometry()
{
	int		status;
	int		gx, gy;
	unsigned int	gwidth, gheight;

	/* determine if there is some geometry to figure out */
	status	= XParseGeometry( geometry, &gx, &gy, &gwidth, &gheight );

	if( status & XValue ) {

		if( status & XNegative )
			mailx	= dwidth  - mail_width  + gx;
		else
			mailx	= gx;
	}
	else
		mailx	= 0;

	if( status & YValue ) {
		
		if( status & YNegative )
			maily	= dheight  - mail_height + gy;
		else
			maily	= gy;
	}
	else
		maily	= 0;

/*	printf( "dwidth %d dheight %d mailx %d maily %d\n", dwidth, dheight, mailx, maily );*/
}

reposition_x_window()
{
	int			x,y,width,height;
	int			ox,oy,owidth,oheight;

	/* determine the sizes for the window according to number of messages */
	/* the size will be centered on the display */
	width			= (max_user_len + max_subject_len + 5) * fwidth;
	height			= (letter_count) * fheight + 15;

	/* add for calculation -- remove for specification */
	width		       += border_width * 2;
	height		       += border_width * 2;

	switch( location ) {
	case	TopCenter:
		x = (dwidth  - width)  / 2;
		y = (dheight - height) / 3;
		break;
	case	NorthEast:
		x = (dwidth - width - COMFORT);
		y = COMFORT;
		break;
	case	NorthWest:
		x = COMFORT;
		y = COMFORT;
		break;
	case	SouthEast:
		x = (dwidth  - width - COMFORT);
		y = (dheight - height - COMFORT);
		break;
	case	SouthWest:
		x = COMFORT;
		y = (dheight - height - COMFORT);
		break;
	}

	/* add for calculation -- remove for specification */
	width		       -= border_width * 2;
	height		       -= border_width * 2;

/*	this is OVERRIDE_REDIRECT -- unneeded 
	xsize.x			= x;
	xsize.y			= y;
	xsize.width		= width;
	xsize.height		= height;
	xsize.min_width		= width;
	xsize.min_height	= height;
	xsize.max_width		= width;
	xsize.max_height	= height;

	xsize.flags		= USPosition | USSize |
				  PPosition  | PSize  |
				  PMinSize   | PMaxSize;

	XSetNormalHints(  display, window, &xsize );
*/

	/* determine if there's a change from last time */
	if( (x != ox) || (y != oy) || (width != owidth) || (height != oheight) ) {
		
		XMoveResizeWindow(display, window, x, y, width, height );

		cwidth	= width;
		cheight	= height;

		ox	= x;
		oy	= y;
		owidth	= width;
		oheight	= height;
	}
}

wait_for_expose( twindow )
Window		twindow;
{
	short			done	= FALSE;

	do {
		XNextEvent( display, &xevent );

		if( xevent.type == Expose && 
		    xevent.xany.window == mwindow )
			repaint_mail_icon();

		if( xevent.type == Expose && 
		    xevent.xany.window == twindow )
			done	= TRUE;
	}
	while( !done );
}	
	
wait_then_unmap()
{
	sleep(wait_time*letter_count);

	XUnmapWindow( display, window );
	flush_events();
}	

flush_events()
{
	/* flush events */
	while( ArePendingEvents() ) {

		XNextEvent( display, &xevent );

		if( xevent.type == Expose && 
		    xevent.xany.window == mwindow )
			repaint_mail_icon();

	}
}


load_xlate_table()
{
	char		xlate_file[STR_MAX];
	char		line[STR_MAX];
	char		*p;
	FILE		*in;

	sprintf( xlate_file, "%s/.xbaff_xlate", getenv("HOME") );

	in 	= fopen( xlate_file, "r" );

	if( in != NULL ) {

		while( fgets( line, sizeof(line), in ) ) {

			if( strchr( line, '~' ) == NULL )
				continue;

			p	= (char*)calloc( strlen(line), 1 );

			line[strlen(line)-1] = '\0';

			strcpy( p, line );

			xlate_name[num_xlate++] = p;
		}

		fclose(in);
	}
}

set_mail_icon()
{
	short	is_new_mail	= (letter_count > 0);

	if( is_new_mail != is_mail ) {

		if( is_new_mail ) {
			toggle_key_led( TRUE );

			XMapRaised( display, mwindow );
			wait_for_expose( mwindow );
		}
		else {
			toggle_key_led( FALSE );

			XUnmapWindow( display, mwindow );
			XFlush( display );
		}

		is_mail = is_new_mail;
	}
}

repaint_mail_icon()
{
	XCopyPlane( display, mail_pixmap, mwindow, regular, 0, 0, 
		  mail_width, mail_height, 0, 0, 1 );
}

/*
 *	change the keyboard led if there's mail / nomail
 */
toggle_key_led( flag )
short		flag;
{
	XKeyboardControl    key;
	
	/* is there one specified? */
	if( keyboard_led == 0 )
		return;

	if( flag )
		key.led_mode = LedModeOn;
	else
		key.led_mode = LedModeOff;
	
	key.led = keyboard_led;

	XChangeKeyboardControl(display, KBLed | KBLedMode, &key);
}

main( argc, argv )
int		argc;
char		**argv;
{
	int		refresh_count	= 0;

	/* load up the mail path - determine the user spool file */
        if ( strcmp(argv[0],"xadvised") == 0 )
        {
           strcpy( file_name, "/usr/mail/advisory" );
           strcpy( def_loc_name, "se" );
        }
        else
        {
           strcpy( file_name, getenv("MAIL") );
           strcpy( def_loc_name, LOC_NAME );
        }

	/* attempt to setup the X world */
	if( init_x_world( argc, argv ) == FALSE )
		exit(1);

	if( strlen ( file_name ) == 0 )
#ifdef SYSV
        	sprintf( file_name, "/usr/mail/%s", getenv("USER") );
#else
        	sprintf( file_name, "/usr/spool/mail/%s", getenv("USER") );
#endif

	/* load translation tables for usernames */
	load_xlate_table();
        sleep(8);

	while(TRUE) {

		current_file_time	= mail_file_time();

		if( current_file_time != last_file_time || 
		    ++refresh_count   > threshhold ) {

			/* get new letter stack */
			if( open_parse_file() ) 
				update_display();
			  
			last_file_time	= current_file_time;
			refresh_count	= 0;
		}

		set_mail_icon();
		flush_events();

		sleep(delay_time);
	}
}
